;
; iface.asm
;
; MAC layer for AX.25 (AX25_MAC)
;
; Parts taken from TFPCX:
; Copyright (C) 1997  Rene Stange, DG0FT, <stange@berlin.snafu.de>
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; 09111998 PE1DNN
;          Modified to make it compilable for MSVC++ 1.5
; 01061998 PE1DNN
;          Added BPQ Receive Callback function

; ***************************************************************************

;       .model  small,c
        .model  small

; *** Equates ***************************************************************

VERSION    = 008h

; *** Externals *************************************************************

        extrn   _reset_intr     : proc
        extrn   _bpq_rx         : proc
        extrn   _ax25_state     : proc
        extrn   _ax25_input     : proc
        extrn   _ax25_output    : proc
        extrn   _ax25_debug     : proc
        extrn   _ax25_bufdbg    : proc

        extrn   _defInt         : byte
        extrn   _l1ports        : byte
        extrn   __psp           : word

; *** Data ******************************************************************

        .data

rand_num dw     0               ; current value

; *** Code ******************************************************************

        .code

; *** AX25_MAC Interface ****************************************************

        public  _iface_mac
_iface_mac label far

        jmp     short L1

        db      'AX25_MAC'

L1:     sti
        push    bx
        push    ds
        push    cx
        push    dx
        push    es
        mov     bx,DGROUP
        mov     ds,bx
        cld

        cmp     ah,1
        je      state
        cmp     ah,2
        je      input
        cmp     ah,3
        je      output
        cmp     ah,0fbh
        je      getpar
        cmp     ah,0fch
        je      bufdbg
        cmp     ah,0fdh
        je      debug
        cmp     ah,0feh
        je      getver
        cmp     ah,0ffh
        je      unload

        mov     ax,0ffffh       ; invalid
        jmp     short exit

state:  pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _ax25_state       ; 'far' call to the state function
        jmp     short exit

input:  pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _ax25_input       ; 'far' call to the input function
        jmp     short exit

output: pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _ax25_output      ; 'far' call to the output function
        jmp     short exit

getpar: mov     al,_l1ports
        jmp     short exit

bufdbg: pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _ax25_bufdbg      ; 'far' call to the bufdbg function
        jmp     short exit

debug:  pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _ax25_debug       ; 'far' call to the output function
        pop     es
        pop     ax              ; dummy, keep dx
        pop     ax              ; dummy, keep cx
        pop     ds
        pop     ax              ; dummy, keep bx
        mov     ax,0
        iret

getver: mov     ax,VERSION
        jmp     short exit

unload: call    _reset_intr
        mov     ax,__psp

exit:   pop     es
        pop     dx
        pop     cx
        pop     ds
        pop     bx
        iret

        public  @call_mac
@call_mac proc near         ; _fastcall

        mov     bl,_defInt
        mov     byte ptr cs:intr,bl
        jmp     short L5        ; clear prefetch queue
L5:     int     0
intr    equ     $-1
        ret

@call_mac endp

; *** Chain *****************************************************************

        public  _chain
_chain proc

        sti
        add     sp,2            ; discard return address
        pop     cx              ; get offset to jump to
        pop     ax              ; get segment
        mov     sp,bp           ; restore SP to point on entry in handler
        xchg    cx,[bp+16]      ; put offset to jump to on stack, get CX
        xchg    ax,[bp+18]      ; put segment, get AX
        pop     es              ; restore saved registers
        pop     ds
        pop     di
        pop     si
        pop     bp
        add     sp,2            ; SP need not to be restored
        pop     bx
        pop     dx
        retf                    ; jump to old handler

_chain endp

; *** Set Stack *************************************************************

        public  _set_stk
_set_stk proc

        pop     bx              ; return address
        pop     ax              ; new stack offset
        pop     cx              ; new stack segment
        mov     dx,ss           ; get old stack segment
        cli
        mov     ss,cx           ; set new stack segment
        xchg    ax,sp           ; set new stack offset and get old one
        sti
        sub     sp,4            ; caller discards arguments from stack
        jmp     bx

_set_stk endp

; *** Random ****************************************************************

    public  _random
_random proc

        mov     ax,rand_num
        test    ah,48h
        jpo     L6
        stc
L6:     rcl     ah,1
        mov     dl,al
        rcr     al,1
        and     ah,7fh
        test    dl,11h
        jpo     L7
        or      ah,80h
L7:     mov     rand_num,ax
        xor     al,ah           ; return value in al
        ret

_random endp

; *** Multi Tasker **********************************************************

    public  _mtasker
_mtasker proc

        mov     ah,30h          ; get DOS version
        int     21h
        cmp     al,20           ; version 20 = OS/2 2.0
        je      L8

        mov     ax,1600h        ; check Windows 3.x Enhanced Mode
        int     2fh
        and     ax,7fh          ; 00h/80h -> no Windows
L8:     ret

_mtasker endp

; *** BPQ RX Callback *******************************************************

    public  _bpq_rx_callback
_bpq_rx_callback proc far
        pushf                   ; push flags so we can use an
                                ; 'interrupt' function in 'C'
        push  cs
        call  _bpq_rx           ; 'far' call to the recieve function
        ret
_bpq_rx_callback endp

; ***************************************************************************

        end
