/*
 *  Copyright (C) 2000-2006  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "digicons.h"
#include "digitype.h"
#include "genlib.h"
#include "read_ini.h"
#include "keep.h"
#include "timer.h"
#include "output.h"

static keep_t *old_data    = NULL;
static keep_t *old_message = NULL;

/*
 * FCS lookup table as generated by fcsgen.c.
 */
static unsigned short fcstab[256] = {
    0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
    0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
    0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
    0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
    0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
    0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
    0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
    0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
    0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
    0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
    0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
    0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
    0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
    0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
    0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
    0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
    0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
    0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
    0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
    0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
    0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
    0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
    0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
    0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
    0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
    0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
    0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
    0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
    0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
    0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
    0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
    0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

#define FCS_START       0xffff  /* Starting bit string for FCS calculation */
#define FCS(fcs, c)     (((fcs) >> 8) ^ fcstab[((fcs) ^ (c)) & 0x00ff])

/* 
 * Compute CCITT CRC-16 over a buffer
 */
static unsigned short calc_crc16(const char *buf, short len)
{
    unsigned short crc;

    crc = FCS_START;

    while (len--)
        crc = FCS(crc, *buf++);

    crc ^= 0xffff;
    return crc;
}

/*
 * Check for data and messages, used to see if data or message is new
 * and has not been handled before.
 *
 * Returns: 0 if the data or message is not new
 *          1 if the data or message is new
 *          2 if the data or message has not been digipeated before
 *            (apperently received directly from the originator, thus new)
 *
 * Bugs: uses 16 bit CRC, may result in false positives. Tradeoff
 *       between storage capacity and accuracy.
 */
short is_new_data_or_message(uidata_t *uidata_p, 
                             char     *data,
                             short     size,
                             keep_t  **keep_store_pp)
{
    keep_t         *old_keep_p;
    keep_t         *keep_p;
    unsigned short  crc;
    short           i;

    /*
     * We have 2 scenarios to distinguish if we saw this data or message
     * before and the data or message is not new.
     *
     * 1) If our call is in the digipeater list with H-Bit set we already
     *    had it and this is not new data or message.
     * 2) If the originator call and CRC matches recent data or message
     *    we assume we already had it too.
     *
     * If the data or message was never digipeated we considder it to be 
     * new data since in that case we are the first digi to receive the data.
     * (in this case the frame will be marked as "LOCAL")
     *
     * First a cleanup is done to remove CRC stamps we not longer need.
     */

    /* cleanup */
    old_keep_p     = *keep_store_pp;
    *keep_store_pp = NULL;
    while(old_keep_p != NULL)
    {
        keep_p     = old_keep_p;
        old_keep_p = old_keep_p->next;

        if(is_elapsed(&(keep_p->keep_timer)) != 0)
        {
            /* this data expired, cleanup */
            free(keep_p);
        }
        else
        {
            /* this data is still valid, add to the rebuild list */
            keep_p->next    = *keep_store_pp;
            *keep_store_pp  = keep_p;
        }
    }

    /*
     * Now before comparing CRC data we first check the digipeater list
     * for our digipeater call.
     */
    for(i = 0; i < uidata_p->ndigi; i++)
    {
        if((uidata_p->digi_flags[i] & H_FLAG) != 0)
        {
            if(strcmp(digi_digi_call, uidata_p->digipeater[i]) == 0)
            {
                /* found our own call with H_BIT set, this is old data */
                return 0;
            }
        }
        else
        {
            /* don't continue the for loop for nothing */
            break;
        }
    }

    /*
     * The list is up to date, now check if we kept the CRC of the
     * reqested data or message, which indicates that we already
     * handled this data or message
     */
    crc = calc_crc16(data, size);

    keep_p = *keep_store_pp;
    while(keep_p != NULL)
    {
        if( (strcmp(keep_p->originator, uidata_p->originator)  == 0)
            &&
            (keep_p->crc == crc)
          )
        {
            /*
             * If we are the first digi it is possibly okay
             * to handle the data, return that it was recoginized but
             * not yet digipeated (retransmission by the originator).
             */
            if(uidata_p->distance == LOCAL)
            {
                return 2;
            }
            /* this is old data */
            return 0;
        }
        keep_p = keep_p->next;
    }

    /* not found, this is new data or a new message */
    return 1;
}

/*
 * Remember the originator and the CRC of the data or message
 * This rememberd call and CRC is used by the 'is_new_data_or_message'
 * function to discover if the data or message is already handled.
 */
void keep_old_data_or_message(short keep_time, uidata_t *uidata_p,
                              char *data, short size, keep_t **keep_store_pp)
{
    keep_t         *keep_p;

    /* We only keep UI data, not AX.25 trafic data */
    if((uidata_p->primitive & ~0x10) != 0x03)
    {
        /* not UI data! */
        vsay("Not UI data, don't keep\r\n");
        return;
    }

    /* If we don't want to remeber previous data then quit here */
    if(keep_time == 0)
    {
        return;
    }

    keep_p = malloc(sizeof(keep_t));
    if(keep_p == NULL)
    {
        vsay("Cannot remember data or message, out of memory\r\n");
        return;
    }

    /* remember the originator and the CRC of the data */
    strcpy(keep_p->originator, uidata_p->originator);
    keep_p->crc = calc_crc16(data, size);

    /* keep "keep_time" seconds */
    start_timer(&(keep_p->keep_timer), keep_time * TM_SEC);

    /* add in front of the list */
    keep_p->next   = *keep_store_pp;
    *keep_store_pp = keep_p;
}

/*
 * Check if originator and data to which 'uidata_p' points is new and
 * not already handled.
 */
short is_new_data(uidata_t *uidata_p)
{
    return is_new_data_or_message(uidata_p, uidata_p->data, 
                                  uidata_p->size, &old_data);
}

/*
 * Remember originator and CRC of the data to which 'uidata_p' points.
 */
void keep_old_data(uidata_t *uidata_p)
{
    short keep_time;

    keep_time = digi_keep_time;

    if( (strchr(digi_data_prefix, uidata_p->data[0]) != NULL)
        &&
        (uidata_p->data[0] != '\0')
      )
    {
        /* data starts with one of the characters             */
        /* in our defined prefix. keep it for a shorter time  */
        keep_time = digi_short_keep_time;
        vsay("First character of data found in 'data_prefix:' array, "
                "keep for shorter time\r\n");
    }

    keep_old_data_or_message(keep_time, uidata_p, uidata_p->data,
                             uidata_p->size, &old_data);
}

/*
 * Check if originator to which 'uidata_p' points and message is new and
 * not already handled.
 */
short is_new_message(char *msg, uidata_t *uidata_p)
{
    return is_new_data_or_message(uidata_p, msg, strlen(msg), &old_message);
}

/*
 * Remember originator to which 'uidata_p' points and CRC of the message. 
 */
void keep_old_message(char *msg, uidata_t *uidata_p)
{
    keep_old_data_or_message(digi_message_keep_time, uidata_p, msg,
                             strlen(msg), &old_message);
}
